"""Functions for generating datasets and annotations."""

from functools import reduce
from typing import Dict

import numpy as np
import pandas as pd

from effiara.preparation import SampleDistributor


def generate_annotator_label(
    true_label: int, correctness_probability: float, num_classes: int
) -> int:
    """Generate an annotator label based on the true label and how correct
       the annotator should be.

    Args:
        true_label (int): the true label of the sample.
        correctness_probability (float): the probability that the annotator
            annotates the sample with the correct label.
        num_classes (int): number of classes

    Returns:
        int: label
    """
    if np.random.rand() < correctness_probability:
        return true_label
    else:
        return np.random.choice(list(set(range(num_classes)) - {true_label}))


def generate_samples(num_samples: int, num_classes: int, seed=None) -> pd.DataFrame:
    """Generate a set of anntotations to be tested in annotator
       reliability assessment. Allows control over how good each
       annotator should be, allowing the assessment of the annotation
       framework.

    Args:
        num_samples (int): the number of samples to generate.
        num_classes (int): the number of possible labels.
        seed (int): the random seed. Default None.

    Returns:
        pd.DataFrame: A dataset of samples.
    """
    if seed is not None:
        np.random.seed(seed)
    true_labels = np.random.randint(0, num_classes, size=num_samples)
    dataset = pd.DataFrame({"true_label": true_labels})

    return dataset


def annotate_samples(
    user_df_dict: Dict[str, pd.DataFrame],
    user_correctness: Dict[str, float],
    num_classes: int,
) -> Dict[str, pd.DataFrame]:
    """Generate annotations according to annotator correctness.

    Args:
        user_df_dict (Dict[str, pd.DataFrame]): dict of usernames: examples.
        user_correctness (Dict[str, pd.DataFrame]): dict from
                                                    usernames to correctness.
        num_classes (int): number of classes.

    Returns:
        Dict[str, pd.DataFrame]: The examples with annotations for each user.
    """

    if set(user_df_dict.keys()) != set(user_correctness.keys()):
        raise ValueError(
            f"Found different users in user_df_dict and user_correctness."
        )  # noqa

    new_user_df_dict = {}
    for user, df in user_df_dict.items():
        # annotate samples
        df.loc[~df["is_reannotation"], f"{user}_label"] = df.loc[
            ~df["is_reannotation"], "true_label"
        ].apply(
            generate_annotator_label,
            correctness_probability=user_correctness[user],
            num_classes=num_classes,
        )
        df.loc[~df["is_reannotation"], f"{user}_confidence"] = 5
        df.loc[~df["is_reannotation"], f"{user}_secondary"] = np.nan

        # reannotations
        df.loc[df["is_reannotation"], f"re_{user}_label"] = df.loc[
            df["is_reannotation"], "true_label"
        ].apply(
            generate_annotator_label,
            correctness_probability=user_correctness[user],
            num_classes=num_classes,
        )
        df.loc[df["is_reannotation"], f"re_{user}_confidence"] = 5
        df.loc[df["is_reannotation"], f"re_{user}_secondary"] = np.nan
        new_user_df_dict[user] = df
    return new_user_df_dict


def consolidate_reannotation(group: pd.DataFrame) -> pd.Series:
    """Helper function to merge a group of rows with the same sample_id.

    Args:
        group (pd.DataFrame): group of dataframes to merge.

    Returns:
        pd.Series: merged row.
    """
    # use first row as base
    base_row = group.iloc[0].copy()

    for col in group.columns:
        if col.startswith("re_"):
            if pd.isna(base_row[col]):
                base_row[col] = (
                    group[col].dropna().iloc[0]
                    if not group[col].dropna().empty
                    else np.nan
                )

    return base_row


def user_df_merge(left: pd.DataFrame, right: pd.DataFrame) -> pd.DataFrame:
    """Merge left and right dataframes on sample_id with checks for
       consistency in shared columns.

    Args:
        left (pd.DataFrame): left dataframe to merge.
        right (pd.DataFrame): right dataframe to merge into left.

    Returns:
        pd.DataFrame: merged dataframe.
    """
    if "is_reannotation" in left.columns:
        left = left.drop(columns=["is_reannotation"])
    if "is_reannotation" in right.columns:
        right = right.drop(columns=["is_reannotation"])
    merged = left.merge(right, on="sample_id", how="outer",
                        suffixes=("", "_dup"))

    # loop over duplicate columns
    for col in merged.columns:
        if col.endswith("_dup"):
            original_col = col.replace("_dup", "")
            if original_col in merged.columns:
                merged[original_col] = merged[original_col].combine_first(
                    merged[col]
                )  # noqa

                # check for consistency
                inconsistent = (
                    (merged[original_col] != merged[col])
                    & merged[original_col].notna()
                    & merged[col].notna()
                )
                if inconsistent.any():
                    raise ValueError(f"Inconsistent values in {original_col}.")

            merged.drop(columns=col, inplace=True)

    merged = (
        merged.groupby("sample_id")
        .apply(lambda group: consolidate_reannotation(group))
        .reset_index(drop=True)
    )

    return merged


def concat_annotations(annotations_dict: Dict[str, pd.DataFrame]):
    return reduce(user_df_merge, annotations_dict.values())


def generate_data(sample_distributor: SampleDistributor,
                  annotator_dict: dict, num_classes: int) -> pd.DataFrame:
    """Generate a set of anntotations to be tested in annotator
       reliability assessment. Allows control over how good each
       annotator should be, allowing the assessment of the annotation
       framework.

    Args:
        sample_distributor (SampleDistributor): sample distributor giving
            the number of annotations in each project.
        annotator_dict (dict): dictionary of annotator and their
            percentage correctness.
        num_classes (int): the number of possible labels.

    Returns:
        pd.DataFrame: The generated dataset.
    """
    dataset = pd.DataFrame()
    num_annotators = len(annotator_dict)
    annotators = list(annotator_dict.keys())
    for i, current_annotator in enumerate(annotators):
        link_1_annotator = annotators[(i + 1) % num_annotators]
        link_2_annotator = annotators[(i + 2) % num_annotators]

        # current annotator single annotations
        true_labels = np.random.randint(
            0, num_classes, size=sample_distributor.single_annotation_project
        )
        data = pd.DataFrame({"true_label": true_labels})

        data[f"{current_annotator}_label"] = data["true_label"].apply(
            generate_annotator_label,
            correctness_probability=annotator_dict[current_annotator],
            num_classes=num_classes,
        )

        # to satisfy current soft label generator
        data[f"{current_annotator}_confidence"] = 5
        data[f"{current_annotator}_secondary"] = np.nan

        # current annotator reannotaions
        reannotation_sample_indices = data.sample(
            sample_distributor.re_annotation_project
        ).index
        data.loc[
            reannotation_sample_indices, f"re_{current_annotator}_label"
        ] = data.loc[reannotation_sample_indices, "true_label"].apply(
            generate_annotator_label,
            correctness_probability=annotator_dict[current_annotator],
            num_classes=num_classes,
        )
        # to satisfy current soft label generator
        data[f"re_{current_annotator}_confidence"] = 5
        data[f"re_{current_annotator}_secondary"] = np.nan

        # add single annotations to dataset
        dataset = pd.concat([dataset, data], ignore_index=True)

        # current and link_1 annotators
        true_labels = np.random.randint(
            0, num_classes, size=sample_distributor.double_annotation_project
        )
        data = pd.DataFrame({"true_label": true_labels})

        data[f"{current_annotator}_label"] = data["true_label"].apply(
            generate_annotator_label,
            correctness_probability=annotator_dict[current_annotator],
            num_classes=num_classes,
        )
        data[f"{link_1_annotator}_label"] = data["true_label"].apply(
            generate_annotator_label,
            correctness_probability=annotator_dict[link_1_annotator],
            num_classes=num_classes,
        )

        # to satisfy current soft label generator
        data[f"{current_annotator}_confidence"] = 5
        data[f"{current_annotator}_secondary"] = np.nan
        data[f"{link_1_annotator}_confidence"] = 5
        data[f"{link_1_annotator}_secondary"] = np.nan

        # add to dataset
        dataset = pd.concat([dataset, data], ignore_index=True)

        # current and link_2 annotators
        true_labels = np.random.randint(
            0, num_classes, size=sample_distributor.double_annotation_project
        )
        data = pd.DataFrame({"true_label": true_labels})

        data[f"{current_annotator}_label"] = data["true_label"].apply(
            generate_annotator_label,
            correctness_probability=annotator_dict[current_annotator],
            num_classes=num_classes,
        )
        data[f"{link_2_annotator}_label"] = data["true_label"].apply(
            generate_annotator_label,
            correctness_probability=annotator_dict[link_2_annotator],
            num_classes=num_classes,
        )

        # to satisfy current soft label generator
        data[f"{current_annotator}_confidence"] = 5
        data[f"{current_annotator}_secondary"] = np.nan
        data[f"{link_2_annotator}_confidence"] = 5
        data[f"{link_2_annotator}_secondary"] = np.nan

        # add to dataset
        dataset = pd.concat([dataset, data], ignore_index=True)

    return dataset



import numpy as np


def is_prob_label(x, num_classes):
    """Check that a value is a soft label (nd vector).

    Args:
        x (np.ndarray): a vector of values.
        num_classes (int): the number of classes.

    Returns:
        bool: whether x is a nd numpy vector.
    """
    is_array = isinstance(x, np.ndarray)
    if not is_array:
        return False
    if np.any(x < 0):
        return False
    is_correct_shape = x.shape == (num_classes,)
    sums_to_one = np.isclose(np.sum(x), 1)
    return is_correct_shape and sums_to_one


def headings_contain_prob_labels(df, heading_1, heading_2, num_classes=3):
    """Check that the given headings contain nothing but probability labels.

    Args:
        df (pd.DataFrame): the dataframe containing data to be worked with.
        heading_1 (str): first heading containing soft labels.
        heading_2 (str): second heading containing soft labels.

    Returns:
        bool: whether headings contain only soft labels.
    """
    checks = df[[heading_1, heading_2]].map(lambda x: is_prob_label(x, num_classes))
    return checks.all(axis=None)


def retrieve_pair_annotations(df, user_x, user_y, suffix="_label"):
    """Get the subset of the dataframe annotated by users
       x and y.

    Args:
        df (pd.DataFrame): the whole dataset.
        user_x (str): name of the user in the form user_x.
        user_x (str): name of the user in the form user_y.
        suffix (str): suffix added to usernames, used to
            find all shared samples.

    Returns:
        pd.DataFrame: copy of the reduced subset containing
                      only samples annotated by both users.
    """
    condition1 = df[f"{user_x}{suffix}"].notna()
    condition2 = df[f"{user_y}{suffix}"].notna()
    return df[condition1 & condition2].copy()


# TODO: this really needs changing to csv_to_list
def csv_to_array(csv_string):
    """Convert csv string (such as value1,value2,...) to
       an array ["value1", "value2", ...]

    Args:
        csv_string (str): string to split and create an array from.

    Returns:
        list[str]: list of strings (if the value passed is a string),
            otherwise returns None.
    """
    if isinstance(csv_string, str):
        split_string = csv_string.split(",")
        return [x.strip() for x in split_string]



"""Functions for computing agreement metrics."""

import krippendorff
import numpy as np
from sklearn.metrics import cohen_kappa_score
from statsmodels.stats.inter_rater import aggregate_raters, fleiss_kappa

from effiara.utils import headings_contain_prob_labels, retrieve_pair_annotations


def pairwise_nominal_krippendorff_agreement(
    pair_df, heading_1, heading_2, label_mapping
):
    """Get the nominal krippendorff agreement between two annotators,
       given two headings for each annotator column containing their
       primary label for each sample.

       Does not require any specific formatting of labels within the columns
       heading_1 and heading_2.

    Args:
        pair_df (pd.DataFrame): dataframe filtered to contain only the samples
                                that allow agreement calculations.
        heading_1 (str): heading of the first column required
                         to calculate agreement.
        heading_2 (str): heading of the second column required
                         to calculate agreement.
        label_mapping (dict): mapping of labels to numeric values.

    Returns:
        float: Krippendorff's Alpha.
    """
    if pair_df[heading_1].isna().any() or pair_df[heading_2].isna().any():
        raise ValueError(
            "One or both of the columns given contain NaN values; the column names may be incorrect or there is an issue with the data."  # noqa
        )

    # convert string categories to numeric values
    pair_df[heading_1 + "_numeric"] = pair_df[heading_1].map(label_mapping)
    pair_df[heading_2 + "_numeric"] = pair_df[heading_2].map(label_mapping)

    if (
        pair_df[heading_1 + "_numeric"].isna().any()
        or pair_df[heading_2 + "_numeric"].isna().any()
    ):
        raise ValueError(
            "Unexpected label found; Please ensure your label mapping is comprehensive."  # noqa
        )

    # turn the pair dataframe into 2d array for Krippendorff calculation
    krippendorff_format_data = (
        pair_df[[heading_1 + "_numeric", heading_2 + "_numeric"]].to_numpy().T
    )
    return krippendorff.alpha(
        reliability_data=krippendorff_format_data, level_of_measurement="nominal"
    )


def pairwise_cohens_kappa_agreement(pair_df, heading_1, heading_2, label_mapping):
    """Cohen's kappa agreement metric between two annotators, given two
       headings for each annotator column containing their primary label
       for each sample.

       Does not require any specific formatting of labels within the columns
       heading_1 and heading_2.

    Args:
        pair_df (pd.DataFrame): dataframe filtered to contain only the samples
                                that allow agreement calculations.
        heading_1 (str): heading of the first column required
                         to calculate agreement.
        heading_2 (str): heading of the second column required
                         to calculate agreement.
        label_mapping (dict): mapping of labels to numeric values.

    Returns:
        float: Cohen's Kappa.
    """
    if pair_df[heading_1].isna().any() or pair_df[heading_2].isna().any():
        raise ValueError(
            "One or both of the columns given contain NaN values; the column names may be incorrect or there is an issue with the data."  # noqa
        )

    if pair_df[heading_1].empty or pair_df[heading_2].empty:
        raise ValueError("One or both of the columns is empty.")

    # convert string categories to numeric values
    pair_df[heading_1 + "_numeric"] = pair_df[heading_1].map(label_mapping)
    pair_df[heading_2 + "_numeric"] = pair_df[heading_2].map(label_mapping)

    if (
        pair_df[heading_1 + "_numeric"].isna().any()
        or pair_df[heading_2 + "_numeric"].isna().any()
    ):
        raise ValueError(
            "Unexpected label found; Please ensure your label mapping is comprehensive."  # noqa
        )

    user_x = pair_df[heading_1 + "_numeric"].to_numpy()
    user_y = pair_df[heading_2 + "_numeric"].to_numpy()

    return cohen_kappa_score(user_x, user_y)


def pairwise_fleiss_kappa_agreement(pair_df, heading_1, heading_2, label_mapping):
    """Fleiss kappa agreement metric between two annotators, given two
       headings for each annotator column containing their primary label
       for each sample.

       Does not require any specific formatting of labels within the columns
       heading_1 and heading_2.

    Args:
        pair_df (pd.DataFrame): dataframe filtered to contain only the samples
                                that allow agreement calculations.
        heading_1 (str): heading of the first column required
                         to calculate agreement.
        heading_2 (str): heading of the second column required
                         to calculate agreement.
        label_mapping (dict): mapping of labels to numeric values.

    Returns:
        float: Fleiss' Kappa.
    """
    if pair_df[heading_1].isna().any() or pair_df[heading_2].isna().any():
        raise ValueError(
            "One or both of the columns given contain NaN values; the column names may be incorrect or there is an issue with the data."  # noqa
        )

    if pair_df[heading_1].empty or pair_df[heading_2].empty:
        raise ValueError("One or both of the columns is empty.")

    # convert string categories to numeric values
    pair_df[heading_1 + "_numeric"] = pair_df[heading_1].map(label_mapping)
    pair_df[heading_2 + "_numeric"] = pair_df[heading_2].map(label_mapping)

    if (
        pair_df[heading_1 + "_numeric"].isna().any()
        or pair_df[heading_2 + "_numeric"].isna().any()
    ):
        raise ValueError(
            "Unexpected label found; Please ensure your label mapping is comprehensive."  # noqa
        )

    fleiss_format_data, _ = aggregate_raters(
        (pair_df[[heading_1 + "_numeric", heading_2 + "_numeric"]].to_numpy())
    )

    return fleiss_kappa(fleiss_format_data, method="fleiss")


def cosine_similarity(vector_a, vector_b):
    """Calculate the cosine similarity between two vectors.

    Args:
        vector_a (np.ndarray)
        vector_b (np.ndarray)

    Returns:
        float: cosine similarity between the two vectors.

    Raises:
        ZeroDivisionError: when vector_a or vector_b is the zero vector.
    """
    if not isinstance(vector_a, np.ndarray) or not isinstance(vector_b, np.ndarray):
        raise ValueError("Vectors a and b MUST be of type np.ndarray.")
    if vector_a.shape != vector_b.shape:
        raise ValueError("Vectors a and b must be of the same shape.")

    numerator = np.dot(vector_a, vector_b)
    divisor = np.linalg.norm(vector_a) * np.linalg.norm(vector_b)

    if np.isclose(divisor, 0):
        raise ZeroDivisionError(
            "Zero division has occurred, vector a or b is the zero vector."
        )

    return numerator / divisor


def pairwise_cosine_similarity(pair_df, heading_1, heading_2, num_classes=3):
    """Calculate the cosine similarity between two columns of soft labels.

       Requires the two headings to be formatted as a soft label
       (list or np.array filled with floats summing to 1).

    Args:
        pair_df (pd.DataFrame): data frame containing annotation data.
        heading_1 (str): heading of first column containing soft labels.
        heading_2 (str): heading of second column containing soft labels.

    Returns:
        float: average cosine similarity between the two sets of soft labels.
    """
    if pair_df[heading_1].isna().any() or pair_df[heading_2].isna().any():
        raise ValueError(
            "One or both of the columns given contain NaN values; the column names may be incorrect or there is an issue with the data."  # noqa
        )

    if len(pair_df[heading_1]) == 0 or len(pair_df[heading_2]) == 0:
        raise ValueError("One or both of the columns is empty.")

    if not headings_contain_prob_labels(pair_df, heading_1, heading_2, num_classes):
        raise ValueError("No probabilistic labels found in dataframe.")

    cosine_similarities = pair_df.apply(
        lambda row: cosine_similarity(row[heading_1], row[heading_2]), axis=1
    )
    return np.sum(cosine_similarities) / len(cosine_similarities)


def calculate_krippendorff_alpha_per_label(
    pair_df, annotator_1_col, annotator_2_col, agreement_type="nominal"
):
    """Calculate Krippendorff's alpha for each label and return the average.

       Requires the data in the given columns to be a binarised array of
       each label (i.e. whether the label is present in the given sample).

    Args:
        annotator_1_col (str): column containing the binarised annotations
            for the first annotator.
        annotator_2_col (str): column containing the binarised annotations
            for the second annotator.
        agreement_type (str): type of agreement:
            - nominal
            - ordinal
            - interval
            - ratio

    Returns:
        float: average Krippendorff's alpha across labels.
    """
    # TODO: check that the column is in the correct format
    annotations_1 = np.vstack(pair_df[annotator_1_col].to_numpy())
    annotations_2 = np.vstack(pair_df[annotator_2_col].to_numpy())

    is_annotations_1_numeric = np.issubdtype(annotations_1.dtype, np.number)
    is_annotations_2_numeric = np.issubdtype(annotations_2.dtype, np.number)
    if not is_annotations_1_numeric or not is_annotations_2_numeric:
        raise ValueError("Annotations are not numeric!")

    if annotations_1.shape != annotations_2.shape:
        raise ValueError("Annotation matrices must have the same shape.")

    alpha_values = []
    # calculate alpha for each label
    for i in range(annotations_1.shape[1]):
        # stack the annotations of both annotators
        label_annotations = np.vstack([annotations_1.T[i], annotations_2.T[i]])

        if len(np.unique(label_annotations)) > 1:
            alpha = krippendorff.alpha(
                reliability_data=label_annotations, level_of_measurement=agreement_type
            )
            alpha_values.append(alpha)
        else:
            alpha_values.append(np.nan)

    # calculate average alpha across labels
    average_alpha = np.nanmean(alpha_values)

    return average_alpha


def pairwise_agreement(
    df,
    user_x,
    user_y,
    label_mapping,
    num_classes,
    metric="krippendorff",
    agreement_type="nominal",
    label_suffix="_label",
):
    """Get the pairwise annotator agreement given the full dataframe.

    Args:
        df (pd.DataFrame): full dataframe containing the whole dataset.
        user_x (str): name of the user in the form user_x.
        user_y (str): name of the user in the form user_y.
        metric (str): agreement metric to use for inter-/intra-annotator agreement.

            * krippendorff: nominal krippendorff's alpha similarity metric on hard labels only.

            * cohen: nominal cohen's kappa similarity metric on hard labels only.

            * fleiss: nominal fleiss kappa similarity metric on hard labels only.

            * multi_krippendorff: krippendorff similarity by label for multilabel classification.

            * cosine: the cosine similarity metric to be used on soft labels.
            
        agreement_type (str): type of agreement.  * nominal

            * ordinal

            * interval

            * ratio

            NOTE: currently only working for multi_krippendorff.
        label_suffix (str): suffix for the label being compared.

    Returns:
        float: agreement between user_x and user_y.

    """
    pair_df = retrieve_pair_annotations(df, user_x, user_y)
    if metric == "krippendorff":
        return pairwise_nominal_krippendorff_agreement(
            pair_df, user_x + label_suffix, user_y + label_suffix, label_mapping
        )
    elif metric == "cohen":
        return pairwise_cohens_kappa_agreement(
            pair_df, user_x + label_suffix, user_y + label_suffix, label_mapping
        )
    elif metric == "fleiss":
        return pairwise_fleiss_kappa_agreement(
            pair_df, user_x + label_suffix, user_y + label_suffix, label_mapping
        )
    elif metric == "cosine":
        return pairwise_cosine_similarity(
            pair_df,
            user_x + label_suffix,
            user_y + label_suffix,
            num_classes=num_classes,
        )
    elif metric == "multi_krippendorff":
        return calculate_krippendorff_alpha_per_label(
            pair_df,
            user_x + label_suffix,
            user_y + label_suffix,
            agreement_type=agreement_type,
        )
    else:
        raise ValueError(f"The metric {metric} was not recognised.")



import pandas as pd
import numpy as np

from effiara.label_generators import LabelGenerator


class OrdinalLabelGenerator(LabelGenerator):

    def __init__(self, annotators: list, label_mapping: dict, label_suffixes=["_label"]):
        super().__init__(annotators, label_mapping)
        self.label_suffixes = label_suffixes

    def add_annotation_prob_labels(self, df: pd.DataFrame) -> pd.DataFrame:
        """Add annotations as an np array [lab1, lab2, ...].

        Args:
            df (pd.DataFrame): dataframe with all annotation
                data to add probability label column to.

        Returns:
            (pd.DataFrame): dataframe with added labels.
        """
        return pd.DataFrame(df.apply(self._add_row_soft_label, axis=1))

    def _add_row_soft_label(self, row) -> pd.Series:
        for annotator in self.annotators:

            cols_to_check = [f"{annotator}{suffix}" for suffix in self.label_suffixes]

            # check cols are all there
            if any(col not in row.index for col in cols_to_check):
                raise ValueError(f"At least one suffix is not in the dataframe for all annotators.")
            
            # add label to non-null
            if row[cols_to_check].notna().all():
                values_list = [self.label_mapping[val] for val in list(row[cols_to_check])]
                row[f"{annotator}_label"] = np.array(values_list, dtype=int)

        return row
    
    def add_sample_prob_labels(self, df: pd.DataFrame, reliability_dict: dict) -> pd.DataFrame:
        return df

    def add_sample_hard_labels(self, df: pd.DataFrame) -> pd.DataFrame:
        return df
    





"""Label generator as in https://doi.org/10.48550/arXiv.2410.14515."""

import numpy as np
import pandas as pd

from . import LabelGenerator


def convert_confidence(confidence, num_classes=3):
    """
    Convert an annotator's confidence value in the range 1-5
    to the range 0-1.

    Args:
        confidence (int): the annotator's confidence in their primary label.

    Returns:
        int: confidence value in the range 0-1
    """
    return (1 / num_classes) + (
        ((num_classes - 1) / num_classes) * ((confidence - 1) / 4)
    )


class EffiLabelGenerator(LabelGenerator):

    def add_annotation_prob_labels(self, df: pd.DataFrame) -> pd.DataFrame:
        """Add soft labels to the dataframe and return the new dataframe.

        Args:
            df (pd.DataFrame): input dataframe.

        Returns:
            pd.DataFrame: dataframe containing generated soft
                          labels for each user.
        """
        return pd.DataFrame(
            df.apply(
                lambda row: self._add_row_soft_labels(
                    row,
                ),
                axis=1,
            )
        )

    def add_sample_prob_labels(self, df, reliability_dict):
        """Generate soft labels for each row in an annotated DataFrame.

           Also generates the sample weightings based on annotator
           reliability scores.

        Args:
            df (pd.DataFrame): input dataframe.

        Returns:
            pd.DataFrame: final dataframe with soft labels and sample weights.
        """
        return pd.DataFrame(
            pd.DataFrame(
                df.apply(
                    lambda row: self._create_row_final_soft_label(
                        row, reliability_dict
                    ),
                    axis=1,
                )
            )
        )

    def add_sample_hard_labels(self, df):
        """Generate hard labels for each row in an annotated DataFrame.

        Args:
            input_df (pd.DataFrame): The annotated DataFrame

        Returns:
            pd.DataFrame: final DataFrame with hard labels.
        """
        return df.apply(lambda row: self._create_row_hard_label(row), axis=1)

    def _create_row_hard_label(self, row):
        """Creates the hard label for the row given there is already
           a soft label in place. This should always be the case.

        Args:
            row (pd.Series): row of the dataframe.

        Returns:
            pd.Series: modified row with the extra "hard_label"
                column.

        Raises:
            ValueError: where no soft label is available for this row -
                this should never be the case.
        """
        soft_label = row["soft_label"]
        soft_label = np.array(soft_label)
        hard_label = np.zeros_like(soft_label)
        # first occurence of max num
        hard_label[np.argmax(soft_label)] = 1
        row["hard_label"] = hard_label
        return row

    def _create_row_final_soft_label(
        self,
        row,
        reliability_dict,
    ):
        """Create final soft label for row with either one or two annotations.
           Adds this final soft label to the row "soft_label" and also
           calculates the final sample weight based on the reliability of the
           annotators that have annotated the sample.

        Args:
            row (pd.Series): single row of the annotation dataframe.
            reliability_dict (dict): dict containing
                                     username: reliability pairs.

        Returns:
            pd.Series: the updated row containing columns of
                       final_soft_label and sample_weight.
        """
        # TODO: convert to list comprehension?
        row_annotators = []
        # loop through each annotator
        for user in self.annotators:
            # check if user has a soft label column that isn't NaN
            if isinstance(row[f"{user}_soft_label"], np.ndarray):
                # if so, append prefix to row_annotators list
                # TOOD: prefix is undefined. What is it?
                row_annotators.append(prefix)

        if len(row_annotators) == 0:
            raise ValueError(
                f"Row number {row.name} does not have annotations, remove this row from the dataset."  # noqa
            )
        elif len(row_annotators) == 1:
            row["soft_label"] = row[f"{row_annotators[0]}_soft_label"]
            row["sample_weight"] = reliability_dict[row_annotators[0]]
        elif len(row_annotators) == 2:
            # weighted average of labels based on annotator reliability
            soft_label_1 = row[f"{row_annotators[0]}_soft_label"]
            reliability_1 = reliability_dict[row_annotators[0]]
            soft_label_2 = row[f"{row_annotators[1]}_soft_label"]
            reliability_2 = reliability_dict[row_annotators[1]]
            reliability_sum = reliability_1 + reliability_2
            avg_rel_1 = reliability_1 / reliability_sum
            avg_rel_2 = reliability_2 / reliability_sum
            row["soft_label"] = avg_rel_1 * soft_label_1 + \
                avg_rel_2 * soft_label_2
            row["sample_weight"] = reliability_sum / 2
        else:
            raise ValueError("This row has more than 2 annotators.")
        return row

    def _add_row_soft_labels(self, row):
        """Add all possible soft labels to the given row.

        Args:
            row (pd.Series): a single row of the full annotation DataFrame.

        Returns:
            row [pd.Series]: Row containing generated soft labels.
        """
        for user in self.annotators:
            valid_user_prefixes = [user, f"re_{user}"]

            for prefix in valid_user_prefixes:
                row[f"{prefix}_soft_label"] = self._create_user_soft_label(
                    row,
                    prefix,
                )

        return row

    def _create_user_soft_label(self, row, user_prefix):
        """Generate the soft label of the given user.

        Args:
            row (pd.Series): a single row of the full annotation DataFrame.
            user_prefix (str): user's prefix in the form username
                               or re_username.

        Returns:
            np.ndarray: vector of soft labels, with each dimension representing
                        probability of the sample fitting the class.
        """

        soft_label = np.zeros(self.num_classes)

        primary_label = row[f"{user_prefix}_label"]

        # user has not made an annotation
        if pd.isna(primary_label):
            return np.nan

        confidence_col = f"{user_prefix}_confidence"

        # get stances and confidence from row
        secondary_label = row[f"{user_prefix}_secondary"]
        primary_confidence = int(row[confidence_col])

        # retrieve index of label
        primary_index = self.label_mapping[primary_label]

        # update primary label with the given confidence
        soft_label[primary_index] = convert_confidence(primary_confidence)

        if pd.isna(secondary_label) or secondary_label == "NotAva":
            # redistribute uniformly
            for i in range(len(soft_label)):
                if i != primary_index:
                    soft_label[i] = (1 - soft_label[primary_index]) / (
                        self.num_classes - 1
                    )
        elif primary_label == secondary_label:
            # edge case where secondary has been changed
            # to be the same label - is equal to all
            # being redistributed to stance 2
            soft_label[primary_index] = 1
        else:
            # redistribute all remaining to stance 2
            secondary_index = self.label_mapping[secondary_label]
            soft_label[secondary_index] = 1 - soft_label[primary_index]

            # if secondary is higher than primary
            if soft_label[secondary_index] > soft_label[primary_index]:
                soft_label[secondary_index] = soft_label[primary_index]
                # case where secondary is bigger than primary,
                # set secondary equal and redistribute rest to other label
                for i in range(len(soft_label)):
                    remaining_probability = (1 - soft_label[primary_index] - soft_label[secondary_index])  # noqa
                    if i != primary_index and i != secondary_index:
                        soft_label[i] = remaining_probability / (self.num_classes - 2)  # noqa

        return soft_label



import numpy as np
import pandas as pd
import pytest

from effiara.data_generator import (
    annotate_samples,
    concat_annotations,
    consolidate_reannotation,
    generate_annotator_label,
    generate_data,
    generate_samples,
    user_df_merge,
)
from effiara.preparation import SampleDistributor


# test generate_annotator_label function
@pytest.mark.parametrize(
    "true_label, correctness_prob, num_classes",
    [
        (0, 1.0, 5),  # Always correct
        (1, 0.0, 5),  # Always wrong, different from true label
        (2, 0.5, 5),  # 50% chance of being correct or incorrect
    ],
)
def test_generate_annotator_label(true_label, correctness_prob, num_classes):
    if correctness_prob == 0.0:
        assert (
            generate_annotator_label(true_label, correctness_prob, num_classes)
            != true_label
        )
    elif correctness_prob == 1.0:
        assert (
            generate_annotator_label(true_label, correctness_prob, num_classes)
            == true_label
        )
    else:
        assert generate_annotator_label(
            true_label, correctness_prob, num_classes
        ) in range(num_classes)


# test generate samples function
def test_generate_samples():
    df = generate_samples(10, 3, seed=42)
    assert isinstance(df, pd.DataFrame)
    assert len(df) == 10
    assert "true_label" in df.columns
    assert df["true_label"].notna().prod() == 1


# test incorrect user input
def test_annotate_samples_invalid_users():
    user_df_dict = {"user_1": pd.DataFrame({"true_label": [0, 1]})}
    user_correctness = {"user_2": 0.8}  # Different user keys
    with pytest.raises(
        ValueError, match="Found different users in user_df_dict and user_correctness"
    ):
        annotate_samples(user_df_dict, user_correctness, 3)


# test annotate samples function
def test_annotate_samples():
    user_df_dict = {
        "user_1": pd.DataFrame({"true_label": [0, 1], "is_reannotation": [False, True]})
    }
    user_correctness = {"user_1": 0.8}
    result = annotate_samples(user_df_dict, user_correctness, 3)
    assert "user_1_label" in result["user_1"].columns
    # only one should be added to regular annotaion column
    assert result["user_1"]["user_1_label"].notna().sum() == 1
    assert "re_user_1_label" in result["user_1"].columns
    # one should be added to reannotation column
    assert result["user_1"]["re_user_1_label"].notna().sum() == 1


# test consolidate_reannotation function
def test_consolidate_reannotation():
    group = pd.DataFrame(
        {
            "sample_id": [1, 1],
            "user_1_label": [1, None],
            "re_user_1_label": [None, 2],
            "is_reannotation": [False, True],
        }
    )
    result = consolidate_reannotation(group)
    assert isinstance(result, pd.Series)
    assert result["sample_id"] == 1
    assert result["user_1_label"] == 1
    assert result["re_user_1_label"] == 2


# test user_df_merge function
def test_user_df_merge():
    df1 = pd.DataFrame({"sample_id": [1, 2], "user_1_label": [0, 1]})
    df2 = pd.DataFrame({"sample_id": [1, 3], "user_2_label": [1, 2]})
    result = user_df_merge(df1, df2)
    assert len(result) == 3
    assert "user_1_label" in result.columns
    assert result["user_1_label"].notna().sum() == 2
    assert "user_2_label" in result.columns
    assert result["user_2_label"].notna().sum() == 2


# test inconsistent values in user_df_merge
def test_user_df_merge_inconsistent_values():
    df1 = pd.DataFrame({"sample_id": [1], "user_1_label": [0]})
    df2 = pd.DataFrame({"sample_id": [1], "user_1_label": [1]})
    with pytest.raises(ValueError, match="Inconsistent values in user_1_label"):
        user_df_merge(df1, df2)


# test concat_annotations function
def test_concat_annotations():
    annotations_dict = {
        "user_1": pd.DataFrame({"sample_id": [1, 2], "user_1_label": [0, 1]}),
        "user_2": pd.DataFrame({"sample_id": [1, 3], "user_2_label": [1, 2]}),
    }
    result = concat_annotations(annotations_dict)
    assert len(result) == 3
    assert "user_1_label" in result.columns
    assert result["user_1_label"].notna().sum() == 2
    assert "user_2_label" in result.columns
    assert result["user_2_label"].notna().sum() == 2


# TODO: update the expected behaviour of small number of annotators
# test generate_data function
def test_generate_data():
    sample_distributor = SampleDistributor(
        annotators=[f"user_{i}" for i in range(1, 7)],  # 6 annotators 1-6
        time_available=10,
        annotation_rate=60,
        num_samples=None,  # 2160
        double_proportion=1.0 / 3,
        re_proportion=0.5,
    )

    sample_distributor.set_project_distribution()
    print(sample_distributor.single_annotation_project)
    print(sample_distributor.double_annotation_project)
    print(sample_distributor.re_annotation_project)
    print(sample_distributor.num_samples)

    annotator_dict = {
        "user_1": 0.8,
        "user_2": 0.7,
        "user_3": 0.4,
        "user_4": 0.9,
        "user_5": 0.95,
        "user_6": 0.75,
    }
    result = generate_data(sample_distributor, annotator_dict, 3)
    assert isinstance(result, pd.DataFrame)
    assert "true_label" in result.columns
    for i in range(1, 7):
        assert f"user_{i}_label" in result.columns
        assert result[f"user_{i}_label"].notna().sum() == 480



import warnings
from itertools import combinations
from typing import Optional

import matplotlib.pyplot as plt
import networkx as nx
import numpy as np
import pandas as pd
import seaborn as sns

from effiara.agreement import pairwise_agreement
from effiara.label_generators import DefaultLabelGenerator, LabelGenerator
from effiara.utils import retrieve_pair_annotations


class Annotations:
    """Class to hold all annotation information for the EffiARA annotation
    framework. Methods include inter- and intra- annotator agreement
    calculations, as well the overall reliability calculation and other
    utilities.

    Attributes:
        label_generator (effiara.LabelGenerator): label generator to create
            individual annotation labels and soft/hard aggregations.
        annotators (list): list of annotator names.
        num_annotators (int): number of annotators
        label_mapping (dict): label mapping of what is in the dataframe to what
            should be used for agreement/training.
        num_classes (int): number of classes.
        agreement_metric (str): agreement metric to be used.
        agreement_suffix (str): label suffix to get the agreement from (such as
            "_label" as the default).
        agreement_type (str): type of agreement (e.g. nominal, ordinal).
        merge_labels (dict): dict of labels to merge.
    """

    def __init__(
        self,
        df: pd.DataFrame,
        label_generator: Optional[LabelGenerator] = None,
        agreement_metric: str = "krippendorff",
        agreement_suffix: str = "_label",
        agreement_type: str = "nominal",
        overlap_threshold: int = 15,
        merge_labels: Optional[dict] = None,
        reliability_alpha: float = 0.5,
        reannotations=False,
    ):
        """
        Args:
            df
            label_generator (effiara.LabelGenerator)
            annotation_col_suffixes (List[str])
            agreement_metric (str)
            agreement_suffix (str)
            agreement_type (str)
            overlap_threshold (str)
            merge_labels (dict): Optional.
            reliability_alpha (float): control strength of intra-annotator
                agreement in reliability calculations.
        """
        # set instance variables
        self.df = df.copy()

        # ensure label generator set
        if label_generator is None:
            label_generator = DefaultLabelGenerator.from_annotations(self.df)

        self.label_generator = label_generator
        assert self.label_generator is not None
        self.annotators = label_generator.annotators
        self.num_annotators = label_generator.num_annotators
        self.label_mapping = label_generator.label_mapping
        self.num_classes = label_generator.num_classes
        self.annotation_col_suffixes = self.label_generator.label_suffixes

        self.agreement_metric = agreement_metric
        self.agreement_suffix = agreement_suffix
        self.agreement_type = agreement_type

        self.overlap_threshold = overlap_threshold

        self.merge_labels = merge_labels

        self.reliability_alpha = reliability_alpha

        # TODO: add some check for reannotations
        self.reannotations = reannotations  # TODO: replace w/ function to look for re_

        # set in self.calculate_inter_annotator_agreement
        self.overall_inter_annotator_agreement = np.nan

        # merge labels
        self.replace_labels()

        # generate annotation labels
        self.df = self.label_generator.add_annotation_prob_labels(self.df)

        # calculate agreements
        self.G = self.init_annotator_graph()
        self.calculate_intra_annotator_agreement()
        self.calculate_inter_annotator_agreement()
        self.calculate_annotator_reliability(alpha=self.reliability_alpha)

    def __getitem__(self, users):
        if isinstance(users, str):
            # Just one user specified
            item = self.G.nodes()[users]
        else:
            if not isinstance(users, (tuple, list)):
                raise KeyError(str(users))
            if len(users) > 2:
                raise ValueError(f"__getitem__ takes one or two users.")
            item = self.G.edges()[users]
        return item

    def replace_labels(self):
        """Merge labels. Uses find and replace so do not switch labels e.g.
        {"misinfo": ["debunk"], "debunk": ["misinfo", "other"]}.
        """
        if not self.merge_labels:
            return

        # TODO: refactor to remove as many loops
        for replacement, to_replace in self.merge_labels.items():
            for label in to_replace:
                for user in self.annotators:
                    for suffix in self.annotation_col_suffixes:
                        label_col = f"{user}{suffix}"
                        self.df[label_col] = self.df[label_col].replace(
                            label, replacement
                        )  # noqa

                        if self.reannotations:
                            re_label_col = "re_" + label_col
                            self.df[re_label_col] = self.df[re_label_col].replace(
                                label, replacement
                            )

    # TODO: rename the label generator names
    def generate_final_labels_and_sample_weights(self):
        """Generate the final labels and sample weights for the dataframe."""
        self.df = self.label_generator.add_sample_prob_labels(
            self.df, self.get_reliability_dict()
        )

        self.df = self.label_generator.add_sample_hard_labels(self.df)

    def init_annotator_graph(self):
        """Initialise the annotator graph with an initial reliability of 1.
        This means each annotator will initially be weighted equally.
        """
        G = nx.Graph()
        for user in self.annotators:
            G.add_node(user, reliability=1)
        return G

    def normalise_edge_property(self, property):
        """Normalise an edge property to have a mean of 1.

        Args:
            property (str): the name of the edge property to normalise.
        """
        total = sum(edge[property] for _, _, edge in self.G.edges(data=True))
        num_edges = self.G.number_of_edges()

        avg = total / num_edges
        if avg < 0:
            raise ValueError(
                "Mean value must be greater than zero, high agreement/reliability will become low and vice versa."
            )  # noqa

        for _, _, edge in self.G.edges(data=True):
            edge[property] /= avg

    def normalise_node_property(self, property):
        """Normalise a node property to have a mean of 1.

        Args:
            property (str): the name of the node property to normalise.
        """
        total = sum(node[property] for _, node in self.G.nodes(data=True))
        num_nodes = self.G.number_of_nodes()

        avg = total / num_nodes
        if avg < 0:
            raise ValueError(
                "Mean value must be greater than zero, high agreement/reliability will become low and vice versa."
            )  # noqa

        for node in self.G.nodes():
            self.G.nodes[node][property] /= avg

    def calculate_inter_annotator_agreement(self):
        """Calculate the inter-annotator agreement between each
        pair of annotators. Each agreement value will be
        represented on the edges of the graph between nodes
        that are representative of each annotator.
        """
        inter_annotator_agreement_scores = {}
        pairs = combinations(self.annotators, 2)
        for current_annotator, link_annotator in pairs:
            pair_df = retrieve_pair_annotations(
                self.df, current_annotator, link_annotator
            )
            if len(pair_df) >= self.overlap_threshold:
                pair = (current_annotator, link_annotator)
                inter_annotator_agreement_scores[pair] = pairwise_agreement(
                    self.df,
                    current_annotator,
                    link_annotator,
                    self.label_mapping,
                    num_classes=self.num_classes,
                    metric=self.agreement_metric,
                    agreement_type=self.agreement_type,
                    label_suffix=self.agreement_suffix,
                )

        # add all agreement scores to the graph
        for users, score in inter_annotator_agreement_scores.items():
            self.G.add_edge(users[0], users[1], agreement=score)

        # TODO: maybe add alternative way of anntotator agreement?
        # i.e. Krippendorff for all annotations if individual annotator
        # doesn't matter
        self.overall_inter_annotator_agreement = np.mean(
            list(inter_annotator_agreement_scores.values())
        )

    # MINOR CHANGE -- add self.reannotations filter
    # TODO: maybe allow to change default value (np.nan for example)
    def calculate_intra_annotator_agreement(self):
        """Calculate intra-annotator agreement."""
        if self.reannotations:
            for user in self.annotators:
                re_user = f"re_{user}"
                try:
                    self.G.nodes[user]["intra_agreement"] = pairwise_agreement(  # noqa
                        self.df,
                        user,
                        re_user,
                        self.label_mapping,
                        num_classes=self.num_classes,
                        metric=self.agreement_metric,
                        agreement_type=self.agreement_type,
                        label_suffix=self.agreement_suffix,
                    )
                except KeyError:
                    warnings.warn(
                        "Key error for calculating intra-annotator agreement. Setting all intra-annotator agreement values to 1."
                    )  # noqa
                    self.G.nodes[user]["intra_agreement"] = 1
                except Exception as e:
                    self.G.nodes[user]["intra_agreement"] = 1
                    print(e)
        else:
            for user in self.annotators:
                self.G.nodes[user]["intra_agreement"] = 1

    def calculate_avg_inter_annotator_agreement(self):
        """Calculate each annotator's average agreement using
        using a weighted average from the annotators around
        them. The average is weighted by the overall reliability
        score of each annotator.
        """
        for node in self.G.nodes():
            edges = self.G.edges(node, data=True)
            # get weighted avg agreement
            weighted_agreement_sum = 0
            weights_sum = 0
            for _, target, edge in edges:
                weight = self.G.nodes[target]["reliability"]
                weights_sum += weight
                weighted_agreement_sum += weight * edge["agreement"]
            self.G.nodes[node]["avg_inter_agreement"] = (
                weighted_agreement_sum / weights_sum if weights_sum else 0
            )

    def reset_annotator_reliability(self):
        for _, node in self.G.nodes(data=True):
            node["reliability"] = 1.0

    def calculate_annotator_reliability(self, alpha=0.5, epsilon=0.001):
        """Recursively calculate annotator reliability, using
           intra-annotator agreement, inter-annotator agreement,
           or a mixture, controlled by the alpha and beta parameters.
           Alpha and Beta must sum to 1.0.

        Args:
            alpha (float): Default 0.5. Value between 0 and 1 controlling weight of intra-annotator agreement.  # noqa
            beta (float): Default 0.5. Value between 0 and 1, controlling weight of inter-annotator agreement.  # noqa
            epsilon (float): Default 0.001. Controls the maximum change from the last iteration to indicate convergence.  # noqa
        """
        if alpha > 1 or alpha < 0:
            raise ValueError("Alpha must be set such that 0 <= alpha <= 1.")

        beta = 1 - alpha

        # set to 1 here to allow for re-calculation with different alpha
        self.reset_annotator_reliability()

        # keep updating until convergence
        max_change = np.inf
        while abs(max_change) > epsilon:
            # print("Running iteration.")
            previous_reliabilties = {
                node: data["reliability"] for (node, data) in self.G.nodes(data=True)
            }

            # calculate the new inter annotator agreement scores
            self.calculate_avg_inter_annotator_agreement()

            # update reliability
            for _, node in self.G.nodes(data=True):
                intra = node["intra_agreement"]
                inter = node["avg_inter_agreement"]
                rel = float(alpha * intra + beta * inter)
                node["reliability"] = rel
            self.normalise_node_property("reliability")

            # find largest change as a marker
            max_change = max(
                [
                    abs(
                        self.G.nodes[node]["reliability"] - previous_reliabilties[node]
                    )  # noqa
                    for node in self.G.nodes()
                ]
            )

    def get_user_reliability(self, username):
        """Get the reliability of a given annotator.

        Args:
            username (str): username of the annotator.

        Returns:
            float: reliability score of the annotator.
        """
        return self.G.nodes[username]["reliability"]

    def get_reliability_dict(self):
        """Get a dictionary of reliability scores per username.

        Returns:
            dict: dictionary of key=username, value=reliability.
        """
        return {node: self.G.nodes[node]["reliability"] for node in self.G.nodes()}

    def display_annotator_graph(self, legend=False):
        """Display the annotation graph."""
        plt.figure(figsize=(12, 12))
        pos = nx.circular_layout(self.G, scale=0.9)

        node_size = 3000
        nx.draw_networkx_nodes(self.G, pos, node_size=node_size)
        nx.draw_networkx_edges(self.G, pos)
        # Get the usernames.
        labels = {node: node.split("_", maxsplit=1)[-1] for node in self.G.nodes()}
        nx.draw_networkx_labels(
            self.G, pos, labels=labels, font_color="white", font_size=24
        )

        # add inter-annotator agreement to edges
        edge_labels = {
            (u, v): f"{d['agreement']:.3f}" for u, v, d in self.G.edges(data=True)
        }
        nx.draw_networkx_edge_labels(self.G, pos, edge_labels=edge_labels, font_size=24)

        # adjust text pos for intra-annotator agreement
        for node, (x, y) in pos.items():
            if x == 0:
                align = "center"
                if y > 0:
                    y_offset = 0.15
                else:
                    y_offset = -0.15
            elif y == 0:
                align = "center"
                y_offset = 0 if x > 0 else -0.15
            elif x > 0:
                align = "left"
                y_offset = 0.15 if y > 0 else -0.15
            else:
                align = "right"
                y_offset = 0.15 if y > 0 else -0.15

            plt.text(
                x,
                y + y_offset,
                s=f"{self.G.nodes[node]['intra_agreement']:.3f}",
                horizontalalignment=align,
                verticalalignment="center",
                fontdict={"color": "black", "size": 24},
            )

        # legend for reliability
        if legend:
            reliability_scores = {
                node: data["reliability"] for (node, data) in self.G.nodes(data=True)
            }
            texts = [
                f"{node}: {score:.3f}" for (node, score) in reliability_scores.items()
            ]
            reliability_text = "Reliability:\n\n" + "\n".join(texts)
            plt.text(
                0.05,
                0.95,
                reliability_text,
                transform=plt.gca().transAxes,
                horizontalalignment="center",
                verticalalignment="top",
                fontsize=12,
                color="black",
            )

        # plot
        plt.axis("off")
        plt.show()

    def display_agreement_heatmap(
        self,
        annotators: Optional[list] = None,
        other_annotators: Optional[list] = None,
        display_upper=False,
    ):
        """Plot a heatmap of agreement metric values for the annotators.

        If both annotators and other_annotators are specifed, compares
        users in annotators to those in other_annotators. Otherwise,
        compare all project annotators to each other.

        Args:
            annotators (list): Optional.
            other_annotators (list): Optional.

        Returns:
            np.ndarray: A matrix of the data displayed on the graph.
            List[str]: List of annotators in the order of the matrix rows.
        """
        mat = nx.to_numpy_array(self.G, weight="agreement")
        # Put intra-agreements on the diagonal
        intras = nx.get_node_attributes(self.G, "intra_agreement")
        intras = np.array(list(intras.values()))
        mat[np.diag_indices(mat.shape[0])] = intras
        agreements = self.G.nodes(data="avg_inter_agreement")
        if annotators is not None and other_annotators is not None:
            matrows = [
                i for (i, user) in enumerate(self.annotators) if user in annotators
            ]
            matcols = [
                i
                for (i, user) in enumerate(self.annotators)
                if user in other_annotators
            ]
            # If we're comparing two sets of annotators,
            # slice the agreement matrix.
            mat = mat[matrows][:, matcols]
            agreements = zip(annotators, np.mean(mat, axis=1))

        sorted_by_agreement = sorted(
            enumerate(agreements), key=lambda n: n[1][1], reverse=True
        )
        ordered_row_idxs = [i for (i, _) in sorted_by_agreement]
        mat = mat[ordered_row_idxs]

        # We now have two possible cases.
        #  1) annotators and other_annotators == None: We're comparing
        #     each annotator to each other. In this case we'll display
        #     only the lower triangle of the agreement heatmap as the
        #     the upper triangle will be identical to the lower.
        #  2) otherwise, we're comparing two possibly distinct sets of
        #     annotators, so we display the full matrix, with rows and
        #     columns sliced according to the annotators specified.
        sorted_users = [user for (i, (user, agree)) in sorted_by_agreement]
        if other_annotators is None:
            mat = mat[:, ordered_row_idxs]
            # Don't display upper triangle, since its redundant.
            if not display_upper:
                mat[np.triu_indices(mat.shape[0], k=1)] = np.nan
            xlabs = ylabs = sorted_users
        else:
            xlabs = [user for user in self.annotators if user in other_annotators]
            ylabs = sorted_users
        sns.heatmap(mat, annot=True, fmt=".3f", xticklabels=xlabs, yticklabels=ylabs)
        plt.show()
        return mat, sorted_users

    def __str__(self):
        return_string = ""
        for node, attrs in self.G.nodes(data=True):
            return_string += f"Node {node} has the following attributes:\n"
            for attr, value in attrs.items():
                return_string += f"  {attr}: {value}\n"
            return_string += "\n"
        return return_string



import re
from abc import ABC, abstractmethod
from typing import List, Optional

import numpy as np
import pandas as pd


class LabelGenerator(ABC):
    """Abstract class for generation of labels for set of annotations.

    This class should be subclassed for each individual annotation
    project. The subclass should override the following methods:
    :code:`add_annotation_prob_labels`,
    :code:`add_sample_prob_labels`,
    :code:`add_sample_hard_labels`

    That is, create a new file with the following:

    .. code-block:: python

        from effiara import LabelGenerator

        class MyLabelGenerator(LabelGenerator):

            def add_annotation_prob_labels(self, df):
                ...

            def add_sample_prob_labels(self, df, reliability_dict):
                ...

            def add_sample_hard_labels(self, df):
                ...
    """

    @classmethod
    def from_annotations(cls, df: pd.DataFrame, num_classes=None):
        """Initialize from an annotations dataframe. Relies on
        labels being stored in the _label columns.

        Args:
            df (pd.DataFrame): annotations, must contain _label columns.
            num_classes (int): if not None, infer from df.
        """
        # Check for columns with hard labels.
        label_cols = [
            c for c in df.columns if c != "true_label" and c.endswith("_label")
        ]
        if len(label_cols) == 0:
            raise ValueError("No *_label columns found in annotations!")

        # Find annotators from label columns
        user_re = re.compile(r"(re_)?([\w -_]+)_.+")
        annotators = []
        # keep label cols with "re_" prefix for label mapping
        user_label_cols = [col for col in label_cols if not col.startswith("re_")]
        for lc in user_label_cols:
            user_re_match = user_re.match(lc)
            if user_re_match is None:
                raise ValueError(f"Improperly formatted label column '{lc}'")
            # Using an `if` rather than set() keeps the
            # annotators in the same order as the columns.
            username = user_re_match.group(2)
            if username not in annotators:
                annotators.append(username)

        # Create a default label mapping
        labels = df[label_cols].values.flatten()
        labels_set = set(labels[pd.notna(labels)].tolist())

        num_classes = num_classes or len(labels_set)
        if num_classes < len(labels_set):
            raise ValueError(
                f"num_classes ({num_classes}) less than the number of classes found in annotations ({len(labels_set)})!"
            )  # noqa
        # We can't guarantee that all the labels
        # are given in the annotations, so if there's a
        # discrepancy add some placeholder labels.
        if num_classes > len(labels_set):
            num_extra = num_classes - len(labels_set)
            for n in range(num_extra):
                labels_set.add(f"_label{n}")

        label_mapping = dict(zip(labels_set, range(num_classes)))
        return cls(annotators, label_mapping)

    def __init__(
        self,
        annotators: list,
        label_mapping: dict,
        label_suffixes: Optional[List[str]] = None,
    ):
        self.annotators = annotators
        self.num_annotators = len(self.annotators)
        self.label_mapping = label_mapping
        self.num_classes = len(label_mapping)
        if label_suffixes is None:
            self.label_suffixes = ["_label"]

    @abstractmethod
    def add_annotation_prob_labels(self, df: pd.DataFrame) -> pd.DataFrame:
        """Add probability distribution (soft) labels
           to each individual annotation.

        Args:
            df (pd.DataFrame): dataframe with all annotation
                data to add probability label column to.

        Returns:
            (pd.DataFrame): dataframe with added labels.
        """
        pass

    @abstractmethod
    def add_sample_prob_labels(
        self, df: pd.DataFrame, reliability_dict: dict
    ) -> pd.DataFrame:
        """Add probability distribution (soft) labels
           to each individual sample, likely using some
           combination of annotation probability
           labels. Can optionally add a sample_weight
           column to weight samples in training based
           on annotator reliability.

        Args:
            df (pd.DataFrame): dataframe with all annotation
                data to add probability label column to.
            reliability_dict (dict): dict of each annotator and
                their reliability score.

        Returns:
            (pd.DataFrame): dataframe with added labels.
        """
        pass

    @abstractmethod
    def add_sample_hard_labels(self, df) -> pd.DataFrame:
        """Implemented to give each sample a one-hot
           hard label for use in the classification
           task.

        Args:
            df (pd.DataFrame): dataframe with all annotation
                data to add probability label column to.

        Returns:
            (pd.DataFrame): dataframe with added labels.
        """
        pass


class DefaultLabelGenerator(LabelGenerator):
    """The most basic LabelGenerator, with support only for hard labels."""

    def add_annotation_prob_labels(self, df):
        return df

    def add_sample_prob_labels(self, df, reliability_dict):
        return df

    def add_sample_hard_labels(self, df):
        return df



"""Module to handle the preparation for anntotation.
This includes:
* calculating number of samples to annotate
* calculating the distribution of samples
"""

import re
import warnings
from typing import List, Optional

import pandas as pd
from sympy import Eq, solve, symbols
from sympy.core.symbol import Symbol


def sample_without_replacement(
    df: pd.DataFrame, n: int
) -> tuple[pd.DataFrame, pd.DataFrame]:
    """Sample a number of data points without replacement.

    Args:
        df (pd.DataFrame): DataFrame to sample from.
        n (int): number of data points to sample.

    Returns:
        pd.DataFrame: complete dataset with sampled data points
            removed.
        pd.DataFrame: sampled data points.
    """
    if not isinstance(n, int):
        raise TypeError("n must be an integer.")
    if n <= 0:
        raise ValueError("Should not attempt to sample 0 or less from DataFrame.")
    n = min([len(df), n])
    sampled_df = df.sample(n)
    return df.drop(sampled_df.index.to_list()), sampled_df


def get_missing_var(variables: dict) -> Symbol:
    """Find the missing variable given a dict of variables.
       Exactly one variable should be None.

    Args:
        variables (dict): dict of variables {var: value}.

    Returns:
        Symbol: symbol of the missing variable using sympy.

    Raises:
        ValueError: if there are no missing variables or more
            than one missing variable.
    """
    if not isinstance(variables, dict):
        raise TypeError("variables must be of type 'dict'.")
    missing_count = 0
    missing_variable = None

    for var, value in variables.items():
        if value is None:
            missing_variable = var
            missing_count += 1
        if missing_count > 1:
            raise ValueError("variables has more than one missing value.")

    if missing_variable is None:
        raise ValueError("variables does not have any missing values.")

    return missing_variable


class SampleDistributor:
    """
    Attributes:
        annotators (list)
        num_annotators (int)
        time_available (float)
        annotation_rate (float)
        num_samples (int)
        double_proportion (float)
        re_proportion (float)
    """

    def __init__(
        self,
        annotators: Optional[List[str]] = None,
        time_available: Optional[float] = None,
        annotation_rate: Optional[float] = None,
        num_samples: Optional[int] = None,
        double_proportion: Optional[float] = None,
        re_proportion: Optional[float] = None,
    ):
        """
        Args:
            annotators (list)
            num_annotators (int)
            time_available (float)
            annotation_rate (float)
            num_samples (int)
            double_proportion (float)
            re_proportion (float)
        """
        if annotators is None:
            num_annotators = None
        else:
            num_annotators = len(annotators)

        self.annotators = annotators

        self.get_variables(
            num_annotators,
            time_available,
            annotation_rate,
            num_samples,
            double_proportion,
            re_proportion,
        )

        # error raised if more than annotators is None
        if self.annotators is None:
            self.annotators = [f"user_{i}" for i in range(1, self.num_annotators + 1)]

    def _assign_variables(self, variables: dict):
        """Assign class level variables from dict of symbolised
           variables (as defined in 'get_variables').

        Args:
            variables (dict): dict of variables to assign.
        """
        n, t, rho, k, d, r = symbols("n t rho k d r")
        self.num_annotators = int(variables.get(n))
        self.time_available = float(variables.get(t))
        self.annotation_rate = float(variables.get(rho))
        self.num_samples = int(variables.get(k))
        self.double_proportion = float(variables.get(d))
        self.re_proportion = float(variables.get(r))

    def get_variables(self,
                      num_annotators: Optional[int] = None,
                      time_available: Optional[float] = None,
                      annotation_rate: Optional[float] = None,
                      num_samples: Optional[int] = None,
                      double_proportion: Optional[float] = None,
                      re_proportion: Optional[float] = None):
        """Solves the annotation framework equation to find the missing
        variable. Only one of the available arguments should be ommitted.

        Args:
            num_annotators (int): number of annotators available [n].
            time_available (float): time available for each annotator
                (assuming they all have the same time available) [t].
            annotation_rate (float): expected rate of annotation per
                unit time (same unit as time_available) [rho].
            num_samples (int): number of desired samples [k].
            double_proportion (float): proportion of the whole dataset that
                should be double-annotated samples (0 <= n <= 1) [d].
            re_proportion (float): proportion of single-annotated samples
                that should be re-annotated (0 <= n <= 1) [r].
        """
        # define variable symbols
        n, t, rho, k, d, r = symbols("n t rho k d r")

        # define distribution equation
        equation = Eq(k, ((2 * d) + ((1 + r) * (1 - d))) ** (-1) * rho * t * n)

        # set vars
        variables = {
            n: num_annotators,
            t: time_available,
            rho: annotation_rate,
            k: num_samples,
            d: double_proportion,
            r: re_proportion,
        }

        # find missing var to solve for
        missing_variable = get_missing_var(variables)

        solution = solve(equation, missing_variable)

        # substitute values into solution
        variables[missing_variable] = solution[0].subs(
            {k: v for k, v in variables.items() if v is not None}
        )

        self._assign_variables(variables)

    def set_project_distribution(self):
        """Set project distributions once all values have been defined."""
        assert self.num_annotators is not None, "num_annotators must be set"  # noqa
        assert self.num_samples is not None, "num_samples must be set"  # noqa
        assert (
            self.double_proportion is not None
        ), "double_proportion must be set"  # noqa
        assert self.re_proportion is not None, "re_proportion must be set"  # noqa

        self.double_annotation_project = round(
            (self.double_proportion * self.num_samples)
            / (2 * self.num_annotators)  # noqa
        )
        self.single_annotation_project = round(
            ((1 - self.double_proportion) * self.num_samples)
            / self.num_annotators  # noqa
        )
        self.re_annotation_project = round(
            self.re_proportion * self.single_annotation_project
        )

    def create_example_distribution_df(self):
        """Create a simple DataFrame to test sample distribution."""
        assert self.num_samples is not None, "num_samples must be set"

        data = {"sample_number": range(1, self.num_samples * 2 + 1)}
        df = pd.DataFrame(data)

        return df

    def distribute_samples(
        self,
        df: pd.DataFrame,
        save_path: Optional[str] = None,
        all_reannotation: bool = False,
    ):
        """Distribute samples based on sample distributor
           settings.

        Args:
            df (pd.DataFrame): dataframe containing samples with
                each row being a separate sample - using a copy
                is recommended.
            save_path (str): (Optional) If not None, dir path to save
                             all data to. If not supplied, a dict of
                             allocations is returned. Default None.
            all_reannotation (bool): whether re-annotations should be sampled
                from all the user's annotations rather than just single
                annotations. In this case, a double annotation project amount
                is sampled from all their annotations.

        Returns:
            dict: Mapping from usernames to assigned samples.
        """
        assert self.num_samples is not None, "num_samples must be set"
        assert self.annotators is not None, "annotators must be set"
        assert (
            self.double_annotation_project is not None
        ), "double_annotation_project must be set"
        assert (
            self.double_proportion is not None
        ), "double_proportion must be set"  # noqa
        assert (
            self.single_annotation_project is not None
        ), "single_annotation_project must be set"
        assert (
            self.re_annotation_project is not None
        ), "re_annotation_project must be set"

        if len(df) < self.num_samples:
            raise ValueError(
                f"DataFrame does not contain enough samples. len(df) [{len(df)}] < num_samples [{self.num_samples}]."  # noqa
            )

        # add sample_id to allow final dataset compilation
        df["sample_id"] = range(len(df))

        # create annotator dict
        annotations_dict = {user: [] for user in self.annotators}

        for i, current_annotator in enumerate(self.annotators):
            link_1_idx = (i + 1) % self.num_annotators
            link_2_idx = (i + 2) % self.num_annotators
            link_1_annotator = self.annotators[link_1_idx]
            link_2_annotator = self.annotators[link_2_idx]
            re_annotation_samples = None

            # single annotations
            if self.double_proportion < 1:
                df, single_samples = sample_without_replacement(
                    df, self.single_annotation_project
                )
                single_samples["is_reannotation"] = False

                annotations_dict[current_annotator].append(single_samples)

                if not all_reannotation:
                    re_annotation_samples = single_samples.sample(
                        self.re_annotation_project
                    )
                    re_annotation_samples["is_reannotation"] = True
                    annotations_dict[current_annotator].append(
                        re_annotation_samples
                    )  # noqa

            # double annotations
            if self.double_annotation_project > 0:
                df, first_double_samples = sample_without_replacement(
                    df, self.double_annotation_project
                )
                first_double_samples["is_reannotation"] = False

                annotations_dict[current_annotator].append(first_double_samples)  # noqa
                annotations_dict[link_1_annotator].append(first_double_samples)  # noqa

                df, second_double_samples = sample_without_replacement(
                    df, self.double_annotation_project
                )
                second_double_samples["is_reannotation"] = False

                annotations_dict[current_annotator].append(
                    second_double_samples
                )  # noqa
                annotations_dict[link_2_annotator].append(second_double_samples)  # noqa

        if len(df) > 0:
            annotations_dict["left_over"] = [df]

        for user, df_list in annotations_dict.items():
            # concat all user's dataframes
            user_df = pd.concat(df_list, ignore_index=True)
            # sample from all if not from singles
            if all_reannotation and user != "left_over":
                re_annotation_samples = user_df.sample(
                    self.re_annotation_project
                )  # noqa
                re_annotation_samples["is_reannotation"] = True
                user_df = pd.concat(
                    [user_df, re_annotation_samples], ignore_index=True
                )  # noqa
            # save df
            if save_path is not None:
                user_df.to_csv(f"{save_path}/{user}.csv", index=False)
            annotations_dict[user] = user_df

        return annotations_dict

    def __str__(self):
        """String representation of sample distribution."""
        return (
            f"Variables:\n"
            f"num_annotators (n): {self.num_annotators}\n"
            f"time_available (t): {self.time_available}\n"
            f"annotation_rate (rho): {self.annotation_rate}\n"
            f"num_samples (k): {self.num_samples}\n"
            f"double_proportion (d): {self.double_proportion}\n"
            f"re_proportion (r): {self.re_proportion}\n"
            f"double_annotation_project: {self.double_annotation_project}\n"
            f"single_annotation_project: {self.single_annotation_project}\n"
            f"re_annotation_project: {self.re_annotation_project}"
        )

    def output_variables(self):
        """Output all variables."""
        print(self)


class SampleRedistributor(SampleDistributor):

    @classmethod
    def from_sample_distributor(cls, sd: SampleDistributor):
        return cls(
            annotators=sd.annotators,
            #time_available=sd.time_available, # need one missing
            annotation_rate=sd.annotation_rate,
            num_samples=sd.num_samples,
            double_proportion=0.0,
            re_proportion=0.0,
        )

    def distribute_samples(
        self,
        df: pd.DataFrame,
        save_path: Optional[str] = None,
        all_reannotation: bool = False,
    ):
        """Distribute samples based on sample distributor
           settings, avoiding allocating samples to annotators
           that have already annotated it.

        Args:
            df (pd.DataFrame): dataframe containing samples with
                each row being a separate sample. Must contain existing
                annotations in columns of the format {user}_label.
            save_path (str): (Optional) If not None, dir path to save all data to.
                             If not supplied, a dict of allocations is returned.
                             Default None.

        Returns:
        annotations (dict): dict of annotator -> pd.DataFrame with allocations
        """
        assert self.double_proportion == 0.0, "Double annotation not yet supported"
        assert self.re_proportion == 0.0, "Reannotation not yet supported"
        assert self.num_samples is not None, "num_samples must be set"
        assert self.annotators is not None, "annotators must be set"
        assert (
            self.double_annotation_project is not None
        ), "double_annotation_project must be set"
        assert (
            self.double_proportion is not None
        ), "double_proportion must be set"  # noqa
        assert (
            self.single_annotation_project is not None
        ), "single_annotation_project must be set"
        assert (
            self.re_annotation_project is not None
        ), "re_annotation_project must be set"
        assert not all_reannotation, "Reannotation not yet supported"

        if len(df) < self.num_samples:
            raise ValueError(
                f"DataFrame does not contain enough samples. len(df) [{len(df)}] < num_samples [{self.num_samples}]."  # noqa
            )
        # Required by other functions
        if "is_reannotation" not in df.columns:
            df["is_reannotation"] = False

        # to hold allocations
        annotations_dict = {user: [] for user in self.annotators}

        user_re = re.compile(r"(re_)?([\w -_]+)_.+")
        label_cols = [c for c in df.columns if c.endswith("_label")]
        annotator_cols = []
        usernames = []
        for lc in label_cols:
            user_re_match = user_re.match(lc)
            assert user_re_match is not None, "Error initialising user_re_match"
            is_reanno = user_re_match.group(1) is not None
            username = user_re_match.group(2)
            if is_reanno is False and username in self.annotators:
                annotator_cols.append(lc)
                usernames.append(username)
        if len(annotator_cols) < 1:
            raise ValueError("No annotations found in dataframe!")

        # First collect the full sample pool for each annotator
        sample_pools = {}
        for ann_col, username in zip(annotator_cols, usernames):
            # nan examples haven't been annotated by this user
            sample_pools[username] = list(df.index[df[ann_col].isna()])

        # Allocate examples round-robin style, which ensures each
        # annotator gets approximately the same number of samples.
        idxs_to_drop = []
        num_failed = 0
        user_idx = 0
        for i, sample in df.iterrows():
            num_users_tried = 0
            while True:
                cur_idx = user_idx % len(annotator_cols)
                ann_col = annotator_cols[cur_idx]
                username = usernames[cur_idx]
                num_users_tried += 1
                user_idx += 1
                if i in sample_pools[username]:
                    annotations_dict[username].append(sample)
                    idxs_to_drop.append(i)
                    break
                if num_users_tried == len(usernames):
                    num_failed += 1
                    break

        for user, annos in annotations_dict.items():
            annotations_dict[user] = pd.DataFrame(annos)

        df.drop(idxs_to_drop, inplace=True)
        if len(df) > 0:
            warnings.warn(
                f"Not all examples were able to be allocated ({len(df)})! Try increasing the number of annotators."
            )  # noqa
            annotations_dict["left_over"] = df

        if save_path is not None:
            for user, user_df in annotations_dict.items():
                user_df.to_csv(f"{save_path}/{user}.csv", index=False)

        return annotations_dict



from typing import List

import numpy as np
import pandas as pd
from sklearn.preprocessing import MultiLabelBinarizer

from effiara.label_generators import LabelGenerator
from effiara.utils import csv_to_array


class TopicLabelGenerator(LabelGenerator):

    def __init__(self, annotators: List[str], label_mapping: dict):
        super().__init__(annotators, label_mapping)
        self.mlb = MultiLabelBinarizer(classes=list(label_mapping.keys()))
        self.mlb.fit([])  # predefined fit so call with empty list

    def binarize(self, label):
        return self.mlb.transform([label])[0]

    def add_annotation_prob_labels(self, df: pd.DataFrame) -> pd.DataFrame:
        """Add binarised labels with a 1 representing a topic's presence."""
        return pd.DataFrame(
            df.apply(
                lambda row: self._add_row_annotation_bin_labels(
                    row,
                ),
                axis=1,
            )
        )

    def _add_row_annotation_bin_labels(self, row):
        all_prefixes = [
            f"{anno_prefix}{annotator}"
            for anno_prefix in ["", "re_"]
            for annotator in self.annotators
        ]

        for prefix in all_prefixes:
            row[f"{prefix}_label"] = csv_to_array(row[f"{prefix}_label"])
            if isinstance(row[f"{prefix}_label"], list):
                row[f"{prefix}_bin_label"] = self.binarize(row[f"{prefix}_label"])
            else:
                row[f"{prefix}_bin_label"] = np.nan

        if "gold" in row:
            row["gold"] = csv_to_array(row["gold"])
            row["gold"] = self.binarize(row["gold"])

        return row

    def add_sample_prob_labels(
        self, df: pd.DataFrame, reliability_dict: dict
    ) -> pd.DataFrame:
        return pd.DataFrame(
            df.apply(
                lambda row: self._add_row_sample_prob_labels(
                    row,
                    reliability_dict,
                ),
                axis=1,
            )
        )

    def _add_row_sample_prob_labels(self, row: pd.Series, reliability_dict: dict):
        # TODO: maybe add in something to account for reannotations
        # set label
        row["soft_label"] = np.zeros(len(self.label_mapping))
        reliability_sum = 0
        num_sample_annotators = 0

        # go through each annotator (only non-reannotations)
        for prefix in self.annotators:
            # check if this annotator annotate the sample
            if isinstance(row[f"{prefix}_bin_label"], (list, np.ndarray)):
                num_sample_annotators += 1
                reliability_sum += reliability_dict[prefix]
                bin_label = np.array(row[f"{prefix}_bin_label"])
                row["soft_label"] += reliability_dict[prefix] * bin_label

        # clip soft label (min=0, max=1)
        row["soft_label"] = np.clip(row["soft_label"] / reliability_sum, 0, 1)
        row["sample_weight"] = reliability_sum / num_sample_annotators

        return row

    def add_sample_hard_labels(self, df: pd.DataFrame) -> pd.DataFrame:
        return pd.DataFrame(
            df.apply(
                lambda row: self._add_row_hard_label(
                    row,
                ),
                axis=1,
            )
        )

    def _add_row_hard_label(self, row: pd.Series) -> pd.Series:
        row["hard_label"] = np.where(row["soft_label"] >= 0.5, 1, 0)
        return row



import os
from pathlib import Path

from setuptools import find_packages, setup

root_dir = Path(__file__).parent
long_description = (root_dir / "README.md").read_text(encoding="utf-8")


def load_requirements(filename="requirements.txt"):
    base_dir = os.path.dirname(os.path.abspath(__file__))
    req_file = os.path.join(base_dir, filename)
    with open(req_file, "r") as f:
        return f.read().splitlines()


setup(
    name="effiara",
    version="0.1.1",
    description="Package for distributing annotations and calculating annotator agreement/reliability using the EffiARA framework.",  # noqa
    long_description=long_description,
    long_description_content_type="text/markdown",
    author="Owen Cook",
    author_email="owenscook1@gmail.com",
    packages=find_packages(where="src"),
    package_dir={"": "src"},
    install_requires=load_requirements(),
    python_requires=">=3.9",
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: MIT License",
        "Operating System :: OS Independent",
    ],
)



import math

import networkx as nx
import pandas as pd
import pytest

from effiara.annotator_reliability import Annotations


# test initialisation of annotations
def test_annotations_initialization():
    df = pd.DataFrame(
        {
            "user_1_label": [0, 1, 2, 1],
            "user_2_label": [0, 1, 2, None],
            "user_3_label": [None, 1, 2, 0],
            "true_label": [0, 1, 2, 1],
        }
    )

    annotations = Annotations(df)

    assert isinstance(annotations, Annotations)
    assert annotations.annotators == ["user_1", "user_2", "user_3"]
    assert annotations.num_annotators == 3
    assert annotations.agreement_metric == "krippendorff"
    assert annotations.agreement_suffix == "_label"


# test replacing labels
def test_replace_labels():
    df = pd.DataFrame(
        {
            "user_1_label": ["misinfo", "debunk", "other"],
            "user_2_label": ["debunk", "misinfo", "other"],
            "true_label": ["misinfo", "debunk", "other"],
        }
    )

    merge_labels = {"other": ["other", "debunk"]}

    annotations = Annotations(df, merge_labels=merge_labels)
    assert annotations.annotators == ["user_1", "user_2"]
    annotations.replace_labels()

    assert (annotations.df["user_1_label"] == ["misinfo", "other", "other"]).all()
    assert (annotations.df["user_2_label"] == ["other", "misinfo", "other"]).all()


# test graph initialisation
def test_init_annotator_graph():
    df = pd.DataFrame({"user_1_label": [0, 1, 2], "user_2_label": [1, 1, 2]})
    annotations = Annotations(df)

    G = annotations.init_annotator_graph()
    assert isinstance(G, nx.Graph)
    assert set(G.nodes) == {"user_1", "user_2"}
    assert all(G.nodes[node]["reliability"] == 1 for node in G.nodes)


# test inter-annotator agreement
@pytest.mark.parametrize(
    "df, expected_value",
    [
        (
            pd.DataFrame(
                {"user_1_label": [0, 1, 2, 1], "user_2_label": [0, 1, 2, None]}
            ),
            1.0,
        ),
        (
            pd.DataFrame(
                {"user_1_label": [1, 2, 2, 1], "user_2_label": [0, 1, 0, None]}
            ),
            -0.25,
        ),
    ],
)
def test_calculate_inter_annotator_agreement(df, expected_value):
    annotations = Annotations(df, overlap_threshold=3)
    annotations.calculate_inter_annotator_agreement()

    assert math.isclose(annotations.overall_inter_annotator_agreement, expected_value)
    assert ("user_1", "user_2") in annotations.G.edges


# test intra-annotator agreement calculation
@pytest.mark.parametrize(
    "df, expected_value",
    [
        (
            pd.DataFrame(
                {"user_1_label": [0, 1, 2, 1], "re_user_1_label": [0, 1, 2, None]}
            ),
            1.0,
        ),
        (
            pd.DataFrame(
                {"user_1_label": [1, 1, 2, 1, 2], "re_user_1_label": [0, 1, 2, None, 2]}
            ),
            0.631578947368421,
        ),
        (
            pd.DataFrame(
                {
                    "user_1_label": [1, 2, 2, 1, None],
                    "re_user_1_label": [0, 1, 0, None, None],
                    "user_2_label": [1, 1, 2, 2, 2],
                    "re_user_2_label": [1, 1, 2, 2, 2],
                }  # add user 2 to not raise lower than 0 error for normalise
            ),
            -0.25,
        ),
    ],
)
def test_calculate_intra_annotator_agreement(df, expected_value):
    annotations = Annotations(df, reannotations=True, overlap_threshold=3)
    annotations.calculate_intra_annotator_agreement()

    assert math.isclose(
        annotations.G.nodes["user_1"]["intra_agreement"], expected_value
    )


# test reliability calculation
@pytest.mark.parametrize("alpha", [0, 0.5, 1])
def test_calculate_annotator_reliability(alpha):
    df = pd.DataFrame(
        {
            "user_1_label": [0, 1, 2, 1],
            "re_user_1_label": [0, 1, 2, None],
            "user_2_label": [0, 1, 2, None],
            "re_user_2_label": [0, 0, 0, None],
            "user_3_label": [0, 1, 0, 1],
            "re_user_3_label": [None, 0, 1, 0],
        }
    )
    annotations = Annotations(
        df, overlap_threshold=3, reannotations=True, reliability_alpha=alpha
    )

    assert annotations.get_user_reliability(
        "user_1"
    ) > annotations.get_user_reliability("user_2")
    assert annotations.get_user_reliability(
        "user_2"
    ) > annotations.get_user_reliability("user_3")


# test invalid alpha
def test_calculate_annotator_reliability_invalid_alpha():
    df = pd.DataFrame({"user1_label": [0, 1, 2]})
    annotations = Annotations(df)

    with pytest.raises(
        ValueError, match="Alpha must be set such that 0 <= alpha <= 1."
    ):
        annotations.calculate_annotator_reliability(alpha=1.5)


# test dict retrieval
def test_get_reliability_dict():
    df = pd.DataFrame(
        {
            "user_1_label": [0, 1, 2, 1],
            "re_user_1_label": [0, 1, 2, None],
            "user_2_label": [0, 1, 2, None],
            "re_user_2_label": [0, 0, 0, None],
            "user_3_label": [0, 1, 0, 1],
            "re_user_3_label": [None, 0, 1, 0],
        }
    )
    annotations = Annotations(df, overlap_threshold=3, reannotations=True)

    reliability_dict = annotations.get_reliability_dict()
    assert isinstance(reliability_dict, dict)
    assert set(reliability_dict.keys()) == {"user_1", "user_2", "user_3"}
    assert all(isinstance(r, float) for r in reliability_dict.values())
